# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Somme handy functions to pipe data.

These are rather convenient for building the interface, and traversing the
AST and so on…

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    )

import itertools

import deal
import returns.curry

T = TypeVar("T")

chain = itertools.chain
"""Just a shortcut for `.itertools.chain`."""

partial = returns.curry.partial
"""Just a shortcut for `.functools.partial`."""


@deal.pure
def cons(el: T, seq: Iterable[T]) -> Iterator[T]:
    """The original cons.

    Args:
        el: An element.
        seq: Any sequence.

    Returns:
        An iterator with the values.


    Examples:
        >>> list(cons(1, (2, 3)))
        [1, 2, 3]

    .. versionadded:: 1.0.0

    """

    return itertools.chain([el], seq)


@deal.pure
def dcons(seqs: Sequence[Tuple[T, Iterable[T]]]) -> Iterator[Tuple[T, ...]]:
    """Like the original cons but for cons of cons.

    Args:
        seqs: Any sequences of sequences.

    Returns:
        An iterator with the values.

    Examples:
        >>> list(dcons([(1, [2, 3]), (4, [5, 6])]))
        [(1, 2, 3), (4, 5, 6)]

    .. versionadded:: 1.0.0

    """

    return (tuple(cons(el, seq)) for el, seq in seqs)


def dmap(func: Callable[[T], Any], seqs: Sequence[Sequence[T]]) -> Iterator[T]:
    """Applies a function to a sequence of sequences.

    Args:
        func: Any callable.
        seqs: Any sequence of sequences.

    Returns:
        An iterator with the values.

    Examples:
        >>> list(dmap(float.__add__, [(.1, .2), (.3, .4)]))
        [0.3..., 0.7]

    .. versionadded:: 1.0.0

    """

    return (func(*seq) for seq in seqs)


@deal.pure
def first(seq: Sequence[T]) -> Optional[T]:
    """Returns the first element of a sequence.

    Args:
        seq: Any sequence.

    Returns:
        The first value, or None.

    Examples:
        >>> first([1, 2, 3])
        1

        >>> first([])
        None

    .. versionadded:: 1.0.0

    """

    return next(iter(seq), None)


@deal.pure
def compact(seq: Iterable[T]) -> Iterator[T]:
    """ Filters falsy values.

    Args:
        seq: Any sequence.

    Returns:
        An iterator with the values.

    Examples:
        >>> list(compact(["1", 0, "a", None, {}]))
        ['1', 'a']

    .. versionadded:: 1.0.0

    """

    return filter(bool, seq)


@deal.pre(lambda seqs: any(map(lambda x: len(x) > 0, seqs)))
@deal.pre(lambda seqs: len(seqs) > 0)
@deal.pure
def flatten(seqs: Sequence[Sequence[T]]) -> Iterator[T]:
    """Flattens a sequences of sequences.

    Args:
        seqs: Any sequence of sequences.

    Returns:
        An iterator with the values.

    Examples:
        >>> list(flatten([(1, 2), (3, 4)]))
        [1, 2, 3, 4]

        >>> list(flatten(["ab", "cd"]))
        ['a', 'b', 'c', 'd']

    .. versionadded:: 1.0.0

    """

    return chain.from_iterable(seqs)



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Retrieves version/change information.

.. versionadded:: 1.0.0

"""


from __future__ import annotations

import deal
import pkg_resources
from git import Repo


@deal.pure
def this() -> str:
    """Retrives the actual version.

    Returns:
        str: Representing the version.

    Examples:
        >>> version = this()
        >>> major, minor, patch, *rest = version.split(".")
        >>> major.isdecimal()
        True

    .. versionadded:: 1.0.0

    """

    return (
        pkg_resources
        .get_distribution("mantic")
        .version
        )


@deal.pure
def last(repo: str = "") -> str:
    """Retrives the last tagged version.

    Args:
        repo: The git repository path.

    Returns:
        str: Representing the version.

    Examples:
        >>> from pathlib import Path

        >>> repo = Path("./tests/fixtures").resolve()
        >>> last(str(repo))
        '10.0.0'

    .. versionadded:: 1.0.0

    """

    return (
        Repo(repo)
        .git
        .describe("--tags", "--abbrev=0", "--first-parent")
        .split()[0]
        )



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Argument schema.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import Any, Optional

import typic


@typic.klass(frozen = True, slots = True, strict = True)
class Argument:
    """An argument.

    Attributes:
        name: The name of the argument.
        types: The annotations, if any.
        default: The default value, if any.

    Examples:
        >>> Argument("count")
        Argument(name='count', default=None)

        >>> Argument("count", "None")
        Argument(name='count', default='None')

        >>> Argument("count", "1")
        Argument(name='count', default='1')

        >>> Argument("count", 1)
        Argument(name='count', default=1)

    .. versionadded:: 1.0.0

    """

    name: str
    default: Optional[Any] = None



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""PySemver's command/queries.

.. versionadded:: 1.0.0

"""

from . import check_signature  # noqa: F401
from ._build_signatures import BuildSignatures  # noqa: F401
from ._check_deprecated import CheckDeprecated  # noqa: F401
from ._check_version import CheckVersion  # noqa: F401



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Typeclass to parse ast nodes.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import Tuple

import ast

import classes


@classes.typeclass
def to_value(instance) -> str:
    """An ast node to it's value in str.

    Args:
        instance: Type of the node.

    Examples:
        >>> to_value(ast.Attribute(attr = "hey!"))
        'hey!'

        >>> to_value(ast.Call(func = "hey!"))
        'hey!'

        >>> to_value(ast.Constant(value = "hey!"))
        'hey!'

        >>> to_value(ast.Ellipsis())
        '...'

        >>> to_value(ast.Name(id = "hey!"))
        'hey!'

        >>> to_value(ast.NameConstant(value = "hey!"))
        'hey!'

        >>> to_value(ast.Num(n = "hey!"))
        'hey!'

        >>> to_value(ast.Str(s = "hey!"))
        'hey!'

        >>> to_value(ast.List(elts=[ast.Constant(value = "hey!")]))
        ('hey!',)

        >>> to_value(ast.Tuple(elts=[ast.Constant(value = "hey!")]))
        ('hey!',)

        >>> to_value(1)
        '1'

        >>> to_value('1')
        '1'

        >>> to_value(None)
        None

    .. versionadded:: 1.0.0

    """

    ...  # pytype: disable=bad-return-type


@to_value.instance(ast.Expr)
def _from_ast_expr(instance: ast.Expr) -> str:
    return str(instance)


@to_value.instance(ast.Slice)
def _from_ast_slice(instance: ast.Slice) -> str:
    return str(instance)


@to_value.instance(ast.Attribute)
def _from_ast_attribute(instance: ast.Attribute) -> str:
    return instance.attr


@to_value.instance(ast.Call)
def _from_ast_call(instance: ast.Call) -> str:
    return to_value(instance.func)  # type: ignore


@to_value.instance(ast.Constant)
def _from_ast_constant(instance: ast.Constant) -> str:
    return to_value(instance.value)


@to_value.instance(ast.Ellipsis)
def _from_ast_ellipsis(instance: ast.Ellipsis) -> str:
    return "..."


@to_value.instance(type(Ellipsis))
def _from_ellipsis(instance: ast.Ellipsis) -> str:
    return "..."


@to_value.instance(ast.Name)
def _from_ast_name(instance: ast.Name) -> str:
    return instance.id


@to_value.instance(ast.NameConstant)
def _from_ast_name_constant(instance: ast.NameConstant) -> str:
    return to_value(instance.value)


@to_value.instance(ast.Num)
def _from_ast_num(instance: ast.Num) -> str:
    return str(instance.n)


@to_value.instance(ast.Str)
def _from_ast_str(instance: ast.Str) -> str:
    return instance.s


@to_value.instance(ast.List)
def _from_ast_list(instance: ast.List) -> Tuple[ast.Expr, ...]:
    return tuple(to_value(el) for el in instance.elts)  # type: ignore


@to_value.instance(ast.Tuple)
def _from_ast_tuple(instance: ast.Tuple) -> Tuple[ast.Expr, ...]:
    return tuple(to_value(el) for el in instance.elts)  # type: ignore


@to_value.instance(int)
def _from_type(instance: int) -> str:
    return str(instance)


@to_value.instance(None)
def _from_none(instance: None) -> str:
    return str(instance)


@to_value.instance(str)
def _from_str(instance: str) -> str:
    return instance



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""The schema of a task.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import MutableMapping, Tuple

import typic


@typic.klass(frozen = True, slots = True, strict = True)
class Task:
    """Represents a task."""

    iterable: Tuple[str]
    optional: Tuple[str]
    help: MutableMapping[str, Tuple[str, ...]]



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Signature schema.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import Tuple

import typic

from ._argument import Argument


@typic.klass(frozen = True, slots = True, strict = True)
class Signature:
    """A signature.

    Attributes:
        name: The name of the function.
        file: The file where the function is defined.
        arguments: All the :class:`.Argument` of the function.
        returns: The return values, if any.

    Examples:
        >>> Signature("greet", "file.py")
        Signature(name='greet', file='file.py', arguments=())

        >>> argument = Argument("count", "1")

        >>> Signature("greet", "file.py", argument)
        Traceback (most recent call last):
        typic.constraints.error.ConstraintValueError...

        >>> Signature("greet", "file.py", (argument,))
        Signature(name='greet', file='file.py', arguments=(Argument(name='co...

    .. versionadded:: 1.0.0

    """

    name: str
    file: str
    arguments: Tuple[Argument, ...] = ()

    def __len__(self) -> int:
        return len(self.arguments)



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Views, or interface, of mantic.

Loosely based on the MVC thing, the idea is to have a clear separation of
concerns between the presentation layer and the rest of the package. Views
are mostly repetitive, but it is yet too soon to refactor them.

.. versionadded:: 1.0.0

"""

from ._to_options import to_options  # noqa: F401



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Mantic CLI.

This sub-package contains the command-line interface of :mod:`.mantic`.

Example usage::

    mantic
    mantic check-deprecated
    mantic --help check-version

.. versionadded:: 1.0.0

"""

from mantic import __version__  # noqa: F401

from .__main__ import main  # noqa: F401



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Base elements of views.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

import deal
from rich.layout import Layout
from rich.panel import Panel

import mantic_hypothesis as st
from mantic import utils

st.register(Layout, st.layouts)
st.register(Panel, st.panels)

_grid = utils.partial(Layout, " ", ratio = 2)
#: To fill around the terminal.

_main = utils.partial(Layout, ratio = 5)
#: To render the main cli view.


@deal.pure
def rows(panel: Panel) -> Layout:
    """Splits the terminal horizontally."""

    layout = Layout()

    layout.split_column(_grid(), _main(panel), _grid())

    return layout


@deal.pure
def columns(panel: Layout) -> Layout:
    """Splits the terminal vertically."""

    layout = Layout()

    layout.split_row(_grid(), _main(panel), _grid())

    return layout



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""File parser.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import Any, cast, Generator, Optional, Set, Tuple

import textwrap

import typic

from ..domain import Signature
from ..infra import repo
from ..types import What
from ._build_signatures import BuildSignatures

_this: str = repo.versions.this()
_that: str = repo.versions.last()


@typic.klass(always = True, slots = True, strict = True)
class ParseFiles:
    """Wrapper around the repo and the signature builder.

    Attributes:
        repo: To query files and changes from.
        this: The base revision.
        that: The revision to compare with.
        diff: The list of files changed between ``this`` and ``that``.
        current: ``this`` or ``that``.
        builder: A signature builder.
        signatures: The list of built signatures.

    Args:
        repo: The repo to use, defaults to :class:`.Repo`.
        this: The revision to use, defaults to ``HEAD``.
        that: The revision to compare ``this`` with, defaults to last version.

    Examples:
        >>> parser = ParseFiles(this = "0.3.0", that = "0.2.0")

        >>> parser.diff
        ('.gitignore', '.python-version', '2', 'Makefile', 'noxfile.py', 'p...)

        >>> with parser(what = "this") as parsing:
        ...     list(parsing)
        ...
        [(1, 10), (2, 10), (3, 10), (4, 10), (5, 10), (6, 10), (7, 10), (8,...]

        >>> this = set(parser.signatures)

        >>> with parser(what = "that") as parsing:
        ...     list(parsing)
        ...
        [(1, 9), (2, 9), (3, 9), (4, 9), (5, 9), (6, 9), (7, 9), (8, 9), (9...]

        >>> that = set(parser.signatures)

        >>> with parser(what = "thus") as parsing:
        ...     print("wut!")
        ...
        Traceback (most recent call last):
        AttributeError: 'ParseFiles' object has no attribute 'thus'

        >>> next(iter(this ^ that & this))  # Added functions…
        Signature(name='...', ...

        >>> next(iter(that ^ this & that))  # Removed functions…
        Signature(name='...', ...

    .. versionadded:: 1.0.0

    """

    this: str
    that: str
    current: Optional[str]
    diff: Tuple[str, ...]
    builder: Optional[BuildSignatures]
    signatures: Optional[Tuple[Signature, ...]]

    def __init__(self, *, this: str = _this, that: str = _that) -> None:
        self.this = this
        self.that = that
        self.diff = repo.files.diff(this, that)
        self.current = None
        self.builder = None
        self.signatures = None

    def __call__(self, *, what: What) -> ParseFiles:
        """We try recover the revision (``this`` or ``that``)."""

        # Fails otherwise.
        # pytype: disable=attribute-error
        self.current = self.__getattribute__(what)
        # pytype: enable=attribute-error

        # And we return ourselves.
        return self

    def __enter__(self) -> Generator[Tuple[int, ...], None, None]:
        # We recover the python files corresponding to ``revison``.
        files: Set[str] = {
            file
            for file in repo.files.tree(cast(str, self.current))
            if file.endswith(".py")
            }

        # We only keep the files changed between ``this`` and ``that``.
        to_parse: Set[str] = files & set(self.diff)

        # We create a builder with the selected files.
        self.builder = BuildSignatures(tuple(to_parse))

        # And finally we iterate over the files…
        for file in self.builder.files:

            # We recover the contents of ``file`` at ``revision``.
            content: str = repo.files.show(cast(str, self.current), file)

            # We sanitize the source code.
            source: str = textwrap.dedent(content)

            # Then pass it on to the signature builder.
            self.builder(source)

            # And we yield a counter to keep the user updated.
            yield self.builder.count, self.builder.total

    def __exit__(self, *args: Any) -> None:
        # We save the signatures for upstream recovery.
        self.signatures = cast(BuildSignatures, self.builder).signatures



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""CLI theme.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from enum import Enum


class ThemeCommon(str, Enum):
    """Common theme for the whole cli."""

    FAIL = "red"
    INFO = "cyan"
    OKAY = "green"
    TEXT = "white"
    WARN = "yellow"
    WORK = "magenta"


class ThemeConsole(str, Enum):
    """Specific rendering theme for instruction screens."""

    BORDER = "cyan"
    HEADER = "magenta"
    ROW = "cyan"
    TITLE = "bold green"



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Version checker.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import Any, cast, Optional, Set, Tuple, TypeVar

import typic

from mantic.domain import Signature, VersionInt, VersionStr

from ..types import What
from ._bump_version import BumpVersion
from ._parse_files import ParseFiles
from .check_signature import CheckSignature

T = TypeVar("T", bound = "CheckVersion")

PARSER = ParseFiles(this = "HEAD")


@typic.klass(always = True, slots = True, strict = True)
class CheckVersion:
    """Checks if the current version is acceptable.

    Attributes:
        bar: A progress bar.
        exit: An exit code.
        parser: A file parser.
        bump_version: A version bump_version.

    .. versionadded:: 1.0.0

    """

    logs: Any
    exit: int
    parser: ParseFiles
    ignore: Tuple[str, ...]
    bump_version: BumpVersion

    def __init__(
            self,
            logs: Any,
            ignore: Tuple[str, ...],
            parser: ParseFiles = PARSER) -> None:
        self.logs = logs
        self.ignore = ignore
        self.exit = VersionInt.NONE
        self.parser = parser
        self.bump_version = BumpVersion()

    def __call__(self) -> None:
        """Runs all the checks."""

        this: Set[Signature] = set(self._parse(self.parser, "this"))
        that: Set[Signature] = set(self._parse(self.parser, "that"))
        diff: Set[str] = set(self.parser.diff)

        (
            self
            ._check_files(self.bump_version, diff)
            ._check_funcs(self.bump_version, VersionStr.MINOR, this, that)
            ._check_funcs(self.bump_version, VersionStr.MAJOR, that, this)
            ._check_version_acceptable(self.bump_version)
            .logs.then()
            )

    def _parse(self, parser: ParseFiles, what: What) -> Tuple[Signature, ...]:
        """Updates status while the parser builds signatures."""

        with parser(what = what) as parsing:
            self.logs.info(f"Parsing files from {parser.current}…\n")
            self.logs.init()

            for count, total in parsing:
                self.logs.push(count, total)

            self.logs.wipe()

        return cast(Tuple[Signature, ...], parser.signatures)

    def _check_files(self: T, bump_version: BumpVersion, files: Set[str]) -> T:
        """Requires a bump if there's a diff in files."""

        what: int = 1
        total: int = len(files)

        self.logs.info("Checking for functional changes…\n")
        self.logs.init()

        for count, file in enumerate(files):
            if not self._is_functional(file):
                continue

            bump_version(what)
            self.exit = bump_version.required
            self.logs.wipe()
            self.logs.warn(f"{str(bump_version.what(what))} {file}\n")
            self.logs.push(count, total)

        self.logs.wipe()

        return self

    def _check_funcs(
            self: T,
            bump_version: BumpVersion,
            what: VersionStr,
            *files: Set[Signature],
            ) -> T:
        """Requires a bump if there's a diff in functions."""

        # We first do a ``hash`` comparison, so it is still grosso modo.
        diff: Set[Signature] = files[0] ^ files[1] & files[0]
        total: int = len(diff)

        self.logs.info(f"Checking for {what} functions…\n")
        self.logs.init()

        for count, this in enumerate(diff):
            name: str
            that: Optional[Signature]
            checker: CheckSignature

            self.logs.push(count, total)

            # If it is not a functional change, we move on.
            if this is None or not self._is_functional(this.file):
                continue

            # We know we will fail already, but we still need to determine
            # the needed version bump.
            self.exit = bump_version.required

            # We will try to find a match between before/after signatures.
            name = this.name
            that = next((that for that in files[1] if that.name == name), None)

            # If we can't find a base signature with the same name, we can just
            # assume the function was added/removed, so minor/major.
            if that is None:
                bump_version(what.to_int())
                self.logs.wipe()
                self.logs.warn(
                    f"{str(bump_version.what(what.to_int()))} "
                    f"{name} => {what.name}\n")
                self.exit = bump_version.required
                continue

            # Now we do a ``small-print`` comparison between signatures.
            f = CheckSignature(this, that)

            if f.score() == bump_version.what(what.to_int()).value:
                bump_version(what.to_int())
                self.logs.wipe()
                self.logs.warn(
                    f"{str(bump_version.what(what.to_int()))} "
                    f"{name}: {f.reason}\n"
                    )
                self.exit = bump_version.required
                continue

        self.logs.wipe()

        return self

    def _check_version_acceptable(self: T, bump_version: BumpVersion) -> T:
        """Requires a bump if there current version is not acceptable."""

        self.logs.info(
            f"Version bump required: {bump_version.required.name}!\n")
        self.logs.okay(f"Current version: {bump_version.this}")

        if bump_version.is_acceptable():
            self.exit = VersionInt.NONE
            return self

        self.logs.fail()

        return self

    def _is_functional(self, file: str) -> bool:
        """Checks if a given ``file`` is whitelisted as functional."""

        return not any(exclude in file for exclude in self.ignore)



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Handy numpy generators.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

import deal
import numpy

import mantic_hypothesis as st

from ..domain import Signature

st.register(Signature, st.signatures)

size_limit = 2e5
"""Just a random size/length sentinel."""

what_limit = 2e20
"""Just a random size/length sentinel."""

limit_bound = deal.chain(
    deal.pre(lambda _: size_limit > _.this >= 0),
    deal.pre(lambda _: size_limit > _.that >= 0),
    deal.pre(lambda _: what_limit > _.what >= 0),
    )
"""Bind these values to some arbitrary limit."""


@deal.pre(lambda this, that: this >= 0 and that >= 0)
@deal.pure
def add(this: int, that: int) -> int:
    """Calculates the offset between two numbers.

    Args:
        this: An upper/lower bound value.
        that: Another upper/lower bound value.

    Returns:
        int: The value.

    Examples:
        >>> add(1, 2)
        1

        >>> add(2, 1)
        0

        >>> add(0, 0)
        0

    .. versionadded:: 1.0.0

    """

    return max(that - this, 0)


@limit_bound
@deal.pure
def pre(this: int, that: int, what: int) -> numpy.ndarray:
    """Pre-populates a numpy array.

    Args:
        this: An upper/lower bound value.
        that: Another upper/lower bound value.
        what: What to pre-populate the array with.

    Returns:
        numpy.ndarray: The pre-populated array.

    Examples:
        >>> pre(10, 5, 1)
        array([1, 1, 1, 1, 1])

    .. versionadded:: 1.0.0

    """

    return numpy.repeat(what, min(this, that))


@limit_bound
@deal.pure
def rep(this: int, that: int, what: int) -> numpy.ndarray:
    """Repeats a value over a numpy array.

    Args:
        this: An upper/lower bound value.
        that: Another upper/lower bound value.
        what: What to repeat over the array with.

    Returns:
        numpy.ndarray: The repeated array.

    Examples:
        >>> rep(10, 5, 1)
        array([1, 1, 1, 1, 1])

    .. versionadded:: 1.0.0

    """

    return numpy.repeat(abs(what), max(add(this, that), add(that, this)))


@limit_bound
@deal.pure
def pop(this: int, that: int, what: int) -> numpy.ndarray:
    """Pre-populates & repeats a value over a numpy array.

    Args:
        this: An upper/lower bound value.
        that: Another upper/lower bound value.
        what: What to populate the array with.

    Returns:
        numpy.ndarray: The populated array.

    Examples:
        >>> pop(10, 5, 1)
        array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

    .. versionadded:: 1.0.0

    """

    return numpy.array([*pre(this, that, what), *rep(this, that, what)])


@limit_bound
@deal.pure
def fill(this: int, that: int, what: int) -> numpy.ndarray:
    """Fill a numpy array with anything.

    Args:
        this: An upper/lower bound value.
        that: Another upper/lower bound value.
        what: What to fill the array with.

    Returns:
        numpy.ndarray: The filled array.

    Examples:
        >>> fill(1, 2, 3)
        array([3, 3])

    .. versionadded:: 1.0.0

    """

    return numpy.array([what] * max(this, that))



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Signature builder.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import (
    Any,
    Callable,
    Generator,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
    )

import ast
import dataclasses
import functools
from pathlib import Path

import deal
import typic

from ..domain import Argument, Signature, Suffix, to_value


def _build_unique_name(
        path: Path,
        node: ast.FunctionDef,
        suffixes: Iterator[Suffix],
        signatures: Tuple[Signature, ...],
        ) -> str:
    """Builds an unique signature name."""

    module: str
    name: str
    decorator: ast.expr

    # We build the module name with the name of the parent path, a
    # folder, and the name of the file, without the extension.
    if len(path.parts) > 1:
        module = f"{path.parts[-2]}.{path.stem}"

    else:
        module = path.stem

    # We compose the name with the name of the module.
    name = f"{module}.{node.name}"

    # We suffix properties, othersise names would duplicate.
    for decorator in node.decorator_list:
        if "property" in ast.dump(decorator):
            name = f"{name}#getter"

        if "setter" in ast.dump(decorator):
            name = f"{name}#setter"

    # Finally we suffix all functions so as to catch the duplicated ones.
    name = f"{name}{next(suffixes).value}"

    # If there are no duplicates, we continue.
    if _is_unique(signatures, name):
        return name

    # Otherwise, we retry…
    return _build_unique_name(path, node, suffixes, signatures)


def _build_posarg(node: ast.FunctionDef) -> Callable[..., Any]:
    """Curryfies the positional arguments builder."""

    return functools.partial(
        _build_argument,
        args = node.args.args,
        defaults = node.args.defaults,
        )


def _build_keyarg(node: ast.FunctionDef) -> Callable[..., Any]:
    """Curryfies the keyword arguments builder."""

    return functools.partial(
        _build_argument,
        args = node.args.kwonlyargs,
        defaults = node.args.kw_defaults,
        )


def _build_argument(
        acc: Tuple[Argument, ...],
        node: ast.arg,
        args: Tuple[Any, ...],
        defaults: Tuple[Any, ...],
        ) -> Tuple[Argument, ...]:
    """Builds an argument.

    Args:
        acc: An accumulated tuple of arguments (from reduce).
        node: An argument node.
        args: A tuple of argument nodes.
        defaults: A tuple of default value nodes.

    Returns:
        A tuple with the arguments.

    Examples:
        >>> acc = []
        >>> node = ast.arg(arg = "arg")
        >>> args = [node]
        >>> defaults = []
        >>> _build_argument(acc, node, args, defaults)
        (Argument(name='arg', default=None),)

    .. versionadded:: 1.0.0

    """

    default: Optional[str]
    argument: Argument

    default = _build_arg_default(len(acc), len(args), defaults)
    argument = Argument(node.arg, default)

    return (*acc, argument)


def _build_arg_default(
        n_acc: int,
        n_arg: int,
        defaults: Tuple[Any, ...],
        ) -> Optional[str]:
    """Builds the default value of an argument.

    Args:
        n_acc: The index of the function argument we're working with.
        n_arg: The total number of a function's arguments.
        defaults: The tuple of a function's default values.

    Returns:
        A default argument value, or None.


    Examples:
        >>> n_acc = 1
        >>> n_arg = 2
        >>> defaults = []
        >>> _build_arg_default(n_acc, n_arg, defaults)
        None

        >>> defaults = [1]
        >>> _build_arg_default(n_acc, n_arg, defaults)
        '1'

        >>> n_acc = 2
        >>> _build_arg_default(n_acc, n_arg, defaults)
        Traceback (most recent call last):
        IndexError: list index out of range

    .. versionadded:: 1.0.0

    """

    n_def: int = len(defaults)
    index: int

    # If there are no default values, we move on.
    if n_def == 0:
        return None

    # Otherwise we would be out of index for defaults.
    if n_arg - n_def > n_acc:
        return None

    # We define the defaults index based on the current visited argument.
    index = n_def + n_acc - n_arg

    return _build(defaults[index])


@deal.raises(NotImplementedError)
@deal.has()
def _build(node: Optional[ast.expr]) -> Any:
    """Generic builder."""

    if node is None:
        return None

    return to_value(node)  # type: ignore


@deal.pure
def _is_unique(seq: Tuple[Signature, ...], name: str) -> bool:
    """Check if a signature's name is unique or not.

    Args:
        seq: A sequence of signatures.
        name: The name of a specific function.

    Returns:
        bool: Is it?


    Examples:
        >>> signature = Signature("name", "file.py")
        >>> seq = [signature]
        >>> _is_unique(seq, "nom")
        True

        >>> _is_unique(seq, "name")
        False

    .. versionadded:: 1.0.0

    """

    is_unique: bool = not next(_where(seq, name), False)

    return is_unique


@deal.pure
def _where(
        seq: Tuple[Signature, ...],
        name: str,
        ) -> Generator[bool, None, None]:
    """Iterates over signatures to find the names named ``name``.

    Args:
        seq: A sequence of signatures.
        name: The name of a specific function.

    Returns:
        A generator to find a signature by name.

    Examples:
        >>> signature = Signature("name", "file.py")
        >>> seq = [signature]
        >>> list(_where(seq, "name"))
        [True]

        >>> list(_where(seq, "nom"))
        []

    .. versionadded:: 1.0.0

    """

    return (True for el in seq if el.name == name)


@typic.klass(always = True, strict = True)
@dataclasses.dataclass
class BuildSignatures(ast.NodeVisitor):
    """Builds signatures from the abstract syntax-tree of a revision.

    Attributes:
        files: The files to build signatures from.
        count: An iteration counter.
        signatures: The built signatures.

    Examples:
        >>> BuildSignatures(["file.py"])
        BuildSignatures(files=['file.py'], count=0, signatures=())

    .. versionadded:: 1.0.0

    """

    files: Union[List[str], Tuple[str, ...]]
    count: int = 0
    signatures: Tuple[Signature, ...] = ()

    @property
    def total(self) -> int:
        """The total number of files to build signatures from.

        Returns:
            int: The number of files.

        Examples:
            >>> builder = BuildSignatures(["file.py"])
            >>> builder.total
            1

        .. versionadded:: 1.0.0

        """

        return len(self.files)

    def __call__(self, source: str) -> None:
        """Builds all signatures from the passed source code.

        Arguments:
            source: The source code to build signatures from.

        Examples:
            >>> builder = BuildSignatures(["file.py"])
            >>> source = [
            ...     "def function(n: List[int] = [1]) -> int:",
            ...     "    return next(iter(n))",
            ...     ]
            >>> builder("\\n".join(source))
            >>> signature = next(iter(builder.signatures))
            >>> argument = next(iter(signature.arguments))

            >>> builder.signatures
            (Signature(name='...file.function', file='file.py', a...

            >>> signature.name
            '...file.function'

            >>> signature.file
            'file.py'

            >>> signature.arguments
            (Argument(name='n', default=('1',)),)

            >>> argument.name
            'n'

            >>> argument.default
            ('1',)

            >>> builder.count
            1

        .. versionadded:: 1.0.0

        """

        node = ast.parse(source, self.files[self.count], "exec")
        self.visit(node)
        self.count += 1

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        """An :obj:`ast` node visitor."""

        file: str
        path: Path
        name: str
        args: Tuple[ast.arg, ...]
        kwds: Tuple[ast.arg, ...]
        posargs: Tuple[Argument, ...]
        keyargs: Tuple[Argument, ...]
        signature: Signature

        # We look for the corresponding ``file``.
        file = self.files[self.count]

        # We find the absolute path of the file.
        path = Path(file).resolve()

        # We take the node name as a base for checks.
        name = node.name

        # We pass if its a private function.
        if name.startswith("_") and not name.endswith("_"):
            return

        # We pass if it is a special function not in __init__ or __call__.
        if name.startswith("__") and name not in ("__init__", "__call__"):
            return

        # We find a unique name for each signature.
        name = _build_unique_name(path, node, iter(Suffix), self.signatures)

        # We build all positional arguments.
        args = tuple(node.args.args)
        posargs = functools.reduce(_build_posarg(node), args, ())

        # We build all keyword arguments.
        kwds = tuple(node.args.kwonlyargs)
        keyargs = functools.reduce(_build_keyarg(node), kwds, ())

        # We build the signature.
        signature = Signature(name, file, posargs + keyargs)

        # And we add it to the list of signatures.
        self.signatures = self.signatures + (signature,)



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Common types.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

import sys

if sys.version_info >= (3, 8):
    from typing import Literal

else:
    from typing_extensions import Literal


What = Literal["this", "that"]



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""The ``check-version`` task.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

import sys

import invoke

from mantic import actions, infra

from ..config import config
from ._task import Task

_task: Task = Task.transmute({
    "iterable": ("ignore",),
    "optional": ("ignore",),
    "help": {
        "ignore": (
            "Paths to ignore",
            f"{', '.join(config.ignore)}",
            ),
        },
    })


@invoke.task(**_task.primitive())
def check_version(_context, ignore):
    """Check if the actual version is valid."""

    if len(ignore) == 0:
        ignore = config.ignore

    task = actions.CheckVersion(infra.logs, tuple(ignore))
    task()
    sys.exit(task.exit.value)



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Exit schema.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from enum import IntEnum


class Exit(IntEnum):
    """An enum with exit codes.

    Examples:
        >>> [exit for exit in Exit]
        [<Exit.OK: 0>, <Exit.KO: 1>]

    .. versionadded:: 1.0.0

    """

    OK = 0
    KO = 1



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Module containing the cli tasks.

.. versionadded:: 1.0.0

"""

from ._tasks import Tasks  # noqa: F401



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Main cli initialiser.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import cast, Tuple

import inspect

import typic
from invoke import Program

import mantic_cli

from .tasks import Tasks
from .views import help_view, home_view, to_options

_tasks: Tasks = Tasks()
#: The list of tasks.

_header: str = "Commands"
#: The home screen header opener.


@typic.klass(always = True, strict = True)
class Main(Program):
    """Interfaces with :mod:`.invoke`."""

    def __init__(self) -> None:
        super().__init__(
            namespace = _tasks,
            version = mantic_cli.__version__,
            )

    def print_help(self) -> None:
        """Intercepts :mod:`.invoke` to print the home screen."""

        return home_view.render(self._make_pairs(_tasks))

    def print_task_help(self, command: str) -> None:
        """Intercepts :mod:`.invoke` to print the help screen."""

        doc: str
        options: Tuple[Tuple[str, ...], ...]

        doc = cast(str, inspect.getdoc(_tasks[command]))
        context = self.parser.contexts[command]
        options = to_options(context.help_tuples())

        return help_view.render(command, doc, options)

    def task_list_opener(self, _command: str = "") -> str:
        """Intercepts :mod:`.invoke` to set the main header name."""

        return _header


main = Main()



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Provides a logger for tasks, with a progress bar!.

.. versionadded:: 1.0.0

"""


from __future__ import annotations

from typing import Sequence

import sys

import deal
import termcolor

limit = 2e5
"""Just a random size/length sentinel."""

_work_icon: str = termcolor.colored("[/]", "cyan")
_info_icon: str = termcolor.colored("[i]", "cyan")
_warn_icon: str = termcolor.colored("[!]", "yellow")
_okay_icon: str = termcolor.colored("[✓]", "green")
_fail_icon: str = termcolor.colored("[x]", "red")

_bar_icon: str = termcolor.colored("|", "green")
_acc_icon: str = termcolor.colored("✓", "green")
_eta_icon: str = termcolor.colored("·", "green")

_bar_size: int = 50


@deal.safe
@deal.has("stdout")
def init() -> None:
    """Initialises the progress bar."""

    sys.stdout.write(_init_message())


@deal.pre(lambda count, total: limit > count >= 0 and limit > total >= 0)
@deal.has("stdout")
def push(count: int, total: int) -> None:
    """Pushes progress to the ``stdout``."""

    done: int = (count + 1) * 100 // total
    sys.stdout.write(_push_message(done))


@deal.safe
@deal.has("stdout")
def okay(message: str) -> None:
    """Prints an okay ``message``."""

    sys.stdout.write(f"{_okay_icon} {message}")


@deal.safe
@deal.has("stdout")
def info(message: str) -> None:
    """Prints an info ``message``."""

    sys.stdout.write(f"{_info_icon} {message}")


@deal.safe
@deal.has("stdout")
def warn(message: str) -> None:
    """Prints a warn ``message``."""

    sys.stdout.write(f"{_warn_icon} {message}")


@deal.safe
@deal.has("stdout")
def fail() -> None:
    """Marks last printed message as failing."""

    sys.stdout.write(f"\r{_fail_icon}")


@deal.safe
@deal.has("stdout")
def then() -> None:
    """Prints a new line and resets the cursor position."""

    sys.stdout.write("\n\r")


@deal.safe
@deal.has("stdout")
def wipe() -> None:
    """Cleans last printed message."""

    sys.stdout.write(_wipe_message())


@deal.pure
def _init_message() -> str:
    message: Sequence[str]
    message = [
        f"{_work_icon} 0%   {_bar_icon}",
        f"{_eta_icon * _bar_size}{_bar_icon}\r",
        ]

    return "".join(message)


@deal.pre(lambda done: limit > done >= 0)
@deal.has()
def _push_message(done: int) -> str:
    message: Sequence[str]
    spaces: str

    spaces = ""
    spaces += [" ", ""][done >= _bar_size * 2]
    spaces += [" ", ""][done >= _bar_size // 5]

    message = [
        f"{_work_icon} {done}% {spaces}{_bar_icon}"
        f"{_acc_icon * (done // 2)}"
        f"{_eta_icon * (_bar_size - done // 2)}"
        f"{_bar_icon}\r"
        ]

    return "".join(message)


@deal.pure
def _wipe_message() -> str:
    return f"{' ' * _bar_size * 3}\r"



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Command-line tools to facilitate semantic versioning.

.. versionadded:: 1.0.0

"""

from . import actions  # noqa: F401
from . import domain  # noqa: F401
from . import infra  # noqa: F401
from . import utils  # noqa: F401

__version__ = infra.repo.versions.this()



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Common types.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import List, Tuple

OptionsAlias = Tuple[Tuple[str, ...], ...]
TupleListAlias = List[Tuple[str, Tuple[str, str]]]
TupleTupleAlias = Tuple[Tuple[str, Tuple[str, str]], ...]


class _TupleListMeta(type):
    """Metaclass to check for a list of tuples."""

    def __instancecheck__(cls, arg: object) -> bool:
        if not isinstance(arg, list):
            return False

        elif not arg[0] or not isinstance(arg[0], (list, tuple)):
            return False

        elif not arg[0][0] or not isinstance(arg[0][0], str):
            return False

        return True


class _TupleTupleMeta(type):
    """Metaclass to check for a tuple of tuples."""

    def __instancecheck__(cls, arg: object) -> bool:
        if not isinstance(arg, tuple):
            return False

        elif not arg[0] or not isinstance(arg[0], (list, tuple)):
            return False

        elif not arg[0][0] or not isinstance(arg[0][0], str):
            return False

        return True


class TupleListType(TupleListAlias, metaclass = _TupleListMeta):
    """A list of tuples."""


class TupleTupleType(TupleTupleAlias, metaclass = _TupleTupleMeta):
    """A tuple of tuples."""



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Check for signature changes and version bump requirements.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import Optional

import dataclasses

import deal
import numpy
import typic

from mantic import utils
from mantic.domain import Signature, VersionInt

limit = 2e5
"""Just a random size/length sentinel."""


@deal.pure
def diff_hash(this: Signature, that: Signature) -> numpy.ndarray:
    """Check if two signatures have a different ``hashes``.

    Args:
        this: A signature.
        that: Another signature.

    Returns:
        numpy.ndarray: An array with the values.

    Examples:
        >>> from mantic.domain import Argument

        >>> argument = Argument("count", default = "0")
        >>> this = Signature("greet", "file.py", (argument,))
        >>> that = Signature("greet", "file.py", (argument,))

        >>> diff_hash(this, that)
        array([0])

        >>> argument = Argument("count")
        >>> this = Signature("greet", "file.py", (argument,))

        >>> diff_hash(this, that)
        array([1])

        >>> diff_hash(that, this)
        array([1])

    .. versionadded:: 1.0.0

    """

    these = utils.fill(len(this), len(that), abs(hash(this)))
    those = utils.fill(len(this), len(that), abs(hash(that)))
    patch = utils.pop(len(this), len(that), VersionInt.PATCH)
    nones = utils.pop(len(this), len(that), VersionInt.NONE)
    return numpy.where(these != those, patch, nones)


@deal.pure
def diff_args(this: Signature, that: Signature) -> numpy.ndarray:
    """Check if two signatures have a diferent arguments.

    Args:
        this: A signature.
        that: Another signature.

    Returns:
        numpy.ndarray: An array with the values.

    Examples:
        >>> from mantic.domain import Argument

        >>> argument = Argument("count", default = "0")
        >>> this = Signature("greet", "file.py", (argument,))
        >>> that = Signature("greet", "file.py", (argument,))

        >>> diff_args(this, that)
        array([0])

        >>> another_argument = Argument("size")
        >>> this = Signature("greet", "file.py", (argument, another_argument))

        >>> diff_args(this, that)
        array([2, 2])

        >>> diff_args(that, this)
        array([3, 3])

        >>> this = Signature("greet", "file.py")

        >>> diff_args(this, that)
        array([3])

        >>> diff_args(that, this)
        array([2])

        >>> that = Signature("greet", "file.py")

        >>> diff_args(this, that)
        array([]...)

        >>> diff_args(that, this)
        array([]...)

    .. versionadded:: 1.0.0

    """

    these = utils.fill(len(this), len(that), len(this))
    those = utils.fill(len(this), len(that), len(that))

    major = utils.pop(len(this), len(that), VersionInt.MAJOR)
    minor = utils.pop(len(this), len(that), VersionInt.MINOR)
    nones = utils.pop(len(this), len(that), VersionInt.NONE)

    conds = [these < those, these > those, True]
    takes = [major, minor, nones]

    return numpy.select(conds, takes)


@deal.pure
def diff_name(this: Signature, that: Signature) -> numpy.ndarray:
    """Check if two signatures have a different argument names.

    Args:
        this: A signature.
        that: Another signature.

    Returns:
        numpy.ndarray: An array with the values.

    Examples:
        >>> from mantic.domain import Argument

        >>> argument = Argument("count")
        >>> this = Signature("greet", "file.py", (argument,))
        >>> that = Signature("greet", "file.py", (argument,))

        >>> diff_name(this, that)
        array([0])

        >>> another_argument = Argument("size")
        >>> this = Signature("greet", "file.py", (argument, another_argument))

        >>> diff_name(this, that)
        array([0, 2])

        >>> diff_name(that, this)
        array([0, 2])

        >>> argument = Argument("counter")
        >>> this = Signature("greet", "file.py", (argument,))

        >>> diff_name(this, that)
        array([3])

        >>> diff_name(that, this)
        array([3])

        >>> this = Signature("greet", "file.py")

        >>> diff_name(this, that)
        array([2])

        >>> diff_name(that, this)
        array([2])

        >>> that = Signature("greet", "file.py")

        >>> diff_name(this, that)
        array([]...)

        >>> diff_name(that, this)
        array([]...)

    .. versionadded:: 1.0.0

    """

    these = numpy.array([a.name for a in this.arguments])
    those = numpy.array([a.name for a in that.arguments])

    # If no arguments, we continue…
    if len(these) == 0 and len(those) == 0:
        return numpy.array([], int)

    # If there are no arguments after, we populate.
    if len(these) == 0:
        these = utils.pop(len(this), len(that), VersionInt.NONE)

    # Inversely as well.
    if len(those) == 0:
        those = utils.pop(len(this), len(that), VersionInt.NONE)

    # We create tuples.
    glued = tuple(zip(these, those))

    # Otherwise we mark missing args after as major.
    conds = [[this != that for this, that in glued], True]
    takes = [
        utils.pre(len(this), len(that), VersionInt.MAJOR),
        utils.pre(len(this), len(that), VersionInt.NONE),
        ]

    # And we mark new args after as minor.
    return numpy.array([
        *numpy.select(conds, takes),
        *utils.rep(len(this), len(that), VersionInt.MINOR),
        ])


@deal.pure
def diff_defs(this: Signature, that: Signature) -> numpy.ndarray:
    """Check if two signatures have a different default values.

    Args:
        this: A signature.
        that: Another signature.

    Returns:
        numpy.ndarray: An array with the values.

    Examples:
        >>> from mantic.domain import Argument

        >>> argument = Argument("count", default = "0")
        >>> this = Signature("greet", "file.py", (argument,))
        >>> that = Signature("greet", "file.py", (argument,))

        >>> diff_args(this, that)
        array([0])

        >>> this = Signature("greet", "file.py")

        >>> diff_args(this, that)
        array([3])

        >>> diff_args(that, this)
        array([2])

        >>> that = Signature("greet", "file.py")

        >>> diff_args(this, that)
        array([]...)

        >>> diff_args(that, this)
        array([]...)

    .. versionadded:: 1.0.0

    """

    major = utils.pop(len(this), len(that), VersionInt.MAJOR)
    minor = utils.pop(len(this), len(that), VersionInt.MINOR)
    nones = utils.pop(len(this), len(that), VersionInt.NONE)

    these = numpy.array([a.default is None for a in this.arguments], int)
    those = numpy.array([a.default is None for a in that.arguments], int)

    if len(these) == 0 and len(those) == 0:
        return numpy.array([], int)

    these.resize((major.size,), refcheck = False)
    those.resize((major.size,), refcheck = False)

    glued = tuple(zip(these, those))

    conds = [
        [this > that for this, that in glued],
        [this < that for this, that in glued],
        True,
        ]

    takes = [
        major,
        minor,
        nones,
        ]

    return numpy.select(conds, takes)


@typic.klass(always = True, slots = True, strict = True)
@dataclasses.dataclass
class CheckSignature:
    """Checks for changes between two :class:`.Signature`.

    Args:
        this: A signature.
        that: Another signature.

    Examples
        >>> from mantic.domain import Argument

        >>> argument = Argument("count")
        >>> this = Signature("greet", "file.py", (argument,))
        >>> that = Signature("greet", "file.py", (argument,))
        >>> service = CheckSignature(this, that)

        >>> service
        CheckSignature(this=Signature(name='greet', file='file.py', argument...

        >>> service.score()
        0

    .. versionadded:: 1.0.0

    """

    this: Signature
    that: Signature
    reason: Optional[str] = None

    def score(self) -> int:
        """Calculates the score, that is the required version bump."""

        hash_score = max(diff_hash(self.this, self.that))
        args_score = max(diff_args(self.this, self.that))
        name_score = max(diff_name(self.this, self.that))
        defs_score = max(diff_defs(self.this, self.that))

        if args_score == 2:
            self.reason = "args-diff"

        if args_score == 3:
            self.reason = "args-diff"

        if name_score == 3:
            self.reason = "args-diff"

        if args_score == 0 and defs_score == 2:
            self.reason = "defaults-diff"

        if args_score == 0 and defs_score == 3:
            self.reason = "defaults-diff"

        if args_score == 2 and defs_score == 0:
            self.reason = "args/defaults-diff"

        return max(hash_score, args_score, name_score, defs_score)



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Version bumber.

.. versionadded:: 1.0.0

"""

from typing import Sequence, Tuple, Type

import re

import deal
import typic

from mantic.infra import repo

from ..domain import VersionInt


@typic.klass(always = True, slots = True, strict = True)
class BumpVersion:
    """Determines the required version bump.

    Attributes:
        this: The actual version.
        that: The last tagged version.
        what: An ``event`` and the associated bump requirement.
        required: The required version bump.

    Examples:
        >>> bumper = BumpVersion()

        >>> bumper.required
        <VersionInt.NONE: 0>

        >>> bumper.what(3)
        <VersionInt.MAJOR: 3>

        >>> bumper(3)
        >>> bumper.required
        <VersionInt.MAJOR: 3>

    .. versionadded:: 1.0.0

    """

    this: str
    that: str
    what: Type[VersionInt]
    required: VersionInt

    @deal.pure
    def __init__(self) -> None:
        self.this = repo.versions.this()
        self.that = repo.versions.last()
        self.what = VersionInt
        self.required = VersionInt.NONE

    @deal.pure
    def __call__(self, bump: int) -> None:
        """Bumps the required version."""

        index = max(self.required.value, VersionInt(bump).value)
        self.required = VersionInt(index)

    @deal.pure
    def is_acceptable(self) -> bool:
        """Determines if the current version is acceptable or not.

        Returns:
            bool:
            True when it is acceptable.
            False otherwise.

        Examples:
            >>> bumper = BumpVersion()
            >>> bumper.is_acceptable()
            True

            >>> bumper(3)
            >>> bumper.this = "1.2.3"
            >>> bumper.that = "1.2.3"
            >>> bumper.is_acceptable()
            False

            >>> bumper.this = "2.0.0"
            >>> bumper.is_acceptable()
            True

            >>> bumper.this = "2.0.0-rc.1+1234"
            >>> bumper.is_acceptable()
            True

            >>> bumper.that = "2.0.0-asdf+1234"
            >>> bumper.is_acceptable()
            True

        .. versionadded:: 1.0.0

        """

        actual_number: int
        actual_is_rel: bool
        before_number: int
        before_is_rel: bool

        # If there's no required bump, we just do not check.
        if self.required == VersionInt.NONE:
            return True

        # We get the actual number and whether it is a release or not.
        actual_number, actual_is_rel = self._extract(self.this)

        # We get the last tagged number and whether it is a release or not.
        before_number, before_is_rel = self._extract(self.that)

        # If both are releases, next version has to be major/minor/patch +1.
        if actual_is_rel and before_is_rel:
            before_number += 1

        # Otherwise we just do not check.
        # It is way too anecdotic for the complexity that the check requires.
        return actual_number >= before_number

    @deal.pure
    def _extract(self, version: str) -> Tuple[int, bool]:
        """Extract a major/minor/patch number from a version string."""

        is_release: bool
        release: str
        rest: Sequence[str]

        # We separate the non-release par, ex: ``-pre+1``.
        release, *rest = re.split("\\+|\\-", version)

        # We get the major/minor/patch number based on the required bump.
        number: str = release.split(".")[3 - self.required.value]

        # Finally we determine if this is a release or not.
        is_release = len(rest) == 0

        return int(number), is_release



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Deprecation checker.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import Any, Iterable, Sequence, Tuple

import ast
import pathlib
import textwrap

import typic

from .. import infra, utils
from ..domain import Exit

_version: str
_version = infra.repo.versions.last()

_files: Tuple[str, ...]
_files = infra.repo.files.tree(_version)


@typic.klass(always = True, strict = True)
class CheckDeprecated(ast.NodeVisitor):
    """Prints the list of features marked as deprecated.

    Attributes:
        count:
            The index of the current ``node`` traversal. Defaults to ``0``.
        exit:
            The exit code for the task handler.
        files:
            The list of files to analyse.
        nodes:
            The corresponding :mod:`ast` of each ``file``.
        version:
            The version to use for the expiration check.

    Args:
        files:
            The list of files to analyse. Defaults to the list of ``.py`` files
            tracked by ``git``.
        version:
            The version to use for the expiration check. Defaults to the
            current version of :mod:`.openfisca_core`.

    .. versionadded:: 1.0.0

    """

    logs: Any
    count: int
    exit: Exit
    files: Sequence[str]
    nodes: Tuple[ast.Module, ...]
    total: int
    version: str
    ignore: Tuple[str, ...]

    def __init__(
            self,
            logs: Any,
            files: Tuple[str, ...] = _files,
            ignore: Tuple[str, ...] = (),
            version: str = _version,
            ) -> None:
        self.logs = logs
        self.exit = Exit.OK
        self.ignore = ignore

        self.files = [
            file for file in files
            if self._is_functional(file) and self._is_python(file)
            ]

        _nodes: Iterable[ast.Module] = (
            self._node(file)
            for file
            in self.files
            )

        self.nodes = tuple(utils.compact(_nodes))
        self.total = len(self.nodes)
        self.version = version

    def __call__(self) -> None:
        """Check fro deprecated features."""

        self.logs.init()

        # We use ``count`` to link each ``node`` with the corresponding
        # ``file``.
        for count, node in enumerate(self.nodes):
            self.count = count
            self.visit(node)
            self.logs.push(self.count, self.total)

        self.logs.wipe()

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        """Defines the ``visit()`` function to inspect the ``node``.

        Args:
            node: The :mod:`ast` node to inspect.

        Attributes:
            keywords:
                The decorator's keywords, see :mod:`.decorators`.
            file:
                The path of a file containing a module.
            path:
                The resolved ``file`` path.
            module:
                The name of the module.
            lineno:
                The line number of each ``node``.
            since:
                The ``since`` keyword's value.
            expires:
                The ``expires`` keyword's value.
            message:
                The message we will print to the user.

        .. versionadded:: 1.0.0

        """

        keywords: Tuple[str, ...]
        file: str
        path: pathlib.Path
        module: str
        lineno: int
        since: str
        expires: str
        message: Sequence[str]

        # We look for the corresponding ``file``.
        file = self.files[self.count]

        # We find the absolute path of the file.
        path = pathlib.Path(file).resolve()

        # We build the module name with the name of the parent path, a
        # folder, and the name of the file, without the extension.
        module = f"{path.parts[-2]}.{path.stem}"

        # We assume the function is defined just one line after the
        # decorator.
        lineno = node.lineno + 1

        for decorator in node.decorator_list:
            # We cast the ``decorator`` to ``callable``.
            if not isinstance(decorator, ast.Call):
                continue

            # We only print out the deprecated functions.
            if "deprecated" not in ast.dump(decorator):
                continue

            # We cast each keyword to ``str``.
            keywords = tuple(
                kwd.value.s
                for kwd in decorator.keywords
                if isinstance(kwd.value, ast.Str)
                )

            # Finally we assign each keyword to a variable.
            since, expires = keywords

            message = [
                f"{module}.{node.name}:{lineno} =>",
                f"Deprecated since: {since}.",
                f"Expiration status: {expires}",
                f"(current: {self.version}).",
                ]

            self.logs.warn(f"{' '.join(message)}")

            # If there is at least one expired deprecation, the handler
            # will exit with an error.
            if self._isthis(expires):
                self.exit = Exit.KO
                self.logs.fail()

            self.logs.then()

    def _isthis(self, version: str) -> bool:
        return self.version == version

    def _node(self, file: str) -> ast.Module:
        source: str

        if pathlib.Path(file).resolve().exists():
            with open(file) as f:
                source = textwrap.dedent(f.read())
                return ast.parse(source, file, "exec")

        return ast.Module()

    def _is_functional(self, file: str) -> bool:
        """Checks if a given ``file`` is whitelisted as functional."""

        return not any(ignore in file for ignore in self.ignore)

    def _is_python(self, file: str) -> bool:
        """Checks if a given ``file`` is a python file."""

        return file.endswith(".py")



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Commons utilities.

.. versionadded:: 1.0.0

"""

from ._arrays import add, fill, pop, pre, rep  # noqa: F401
from ._functions import (  # noqa: F401
    chain,
    compact,
    cons,
    dcons,
    dmap,
    first,
    flatten,
    partial,
    )



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Home screen view.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import Tuple

import deal
from rich.console import Console
from rich.layout import Layout
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

import mantic_hypothesis as st
from mantic import __name__, utils

from .. import __version__
from ._base import columns, rows
from ._theme import ThemeConsole

st.register(Layout, st.layouts)
st.register(Panel, st.panels)
st.register(Table, st.tables)

_headers = "Command", "Description"
#: Main command headers.

_console = Console()
#: Internal function used to render.


@deal.has("stdout")
@deal.safe
def render(tasks: Tuple[Tuple[str, ...], ...]) -> None:
    """Render the home view.

    Args:
        tasks:
            A sequence of tuples containing the commands and their
            descriptions, default values, for each command.

    Examples:
        >>> render([("say-hi!", "HI!!! ❤ ❤ ❤")])
        ...say-hi!...HI!!! ❤ ❤ ❤...

    .. versionadded:: 1.0.0

    """

    content = _content(tasks)
    main = _main(content)
    root = _root(main)

    _console.print(root)


@deal.pure
def _root(main: Panel) -> Layout:
    """Global home container.

    Args:
        main: Main container to wrap.

    Returns:
        Layout: To render.

    Examples:
        >>> content = _content(())
        >>> main = _main(content)
        >>> _root(main)
        Layout()

    .. versionadded:: 1.0.0

    """

    return columns(rows(main))


@deal.pure
def _main(content: Table) -> Panel:
    """The main container, or panel.

    Args:
        content: The inner container.

    Returns:
        The outer panel.

    Examples:
        >>> content = _content(())
        >>> main = _main(content)
        >>> main.title
        ...mantic...

    .. versionadded:: 1.0.0

    """

    return Panel(
        content,
        border_style = ThemeConsole.BORDER,
        padding = 5,
        title = _usage(),
        subtitle = __version__,
        )


@deal.pure
def _content(tasks: Tuple[Tuple[str, ...], ...]) -> Table:
    """Inner usage instructions content.

    Args:
        tasks:
            A sequence of tuples containing the commands and their
            descriptions, default values, for each command.

    Returns:
        The main content.

    Examples:
        >>> content = _content([("say-hi!", "HI!!! ❤ ❤ ❤")])
        >>> list(map(lambda option: option._cells, content.columns))
        [['say-hi!'], ['HI!!! ❤ ❤ ❤']]

    .. versionadded:: 1.0.0

    """

    table = Table(
        box = None,
        padding = (0, 5, 1, 10),
        row_styles = [ThemeConsole.ROW],
        style = ThemeConsole.HEADER,
        )

    list(map(table.add_column, _headers))

    list(utils.dmap(table.add_row, tasks))

    return table


@deal.pure
def _usage() -> Text:
    """Usage instructions.

    Returns:
        A formatted usage text.

    Examples:
        >>> _usage()
        <text 'Usage: mantic [--help] <command> …' ...>

    .. versionadded:: 1.0.0

    """

    name = utils.first(__name__.split("."))
    text = Text(f"Usage: {name} [--help] <command> …")

    text.stylize(ThemeConsole.TITLE)

    return text



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Version schema.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from enum import Enum, IntEnum


class VersionStr(str, Enum):
    """An enum used to explain the required version.

    Examples:
        >>> [version.name for version in VersionStr]
        ['NONE', 'PATCH', 'MINOR', 'MAJOR']

        >>> [version.value for version in VersionStr]
        ['', '~', '+', '-']

    .. versionadded:: 1.0.0

    """

    NONE = ""
    PATCH = "~"
    MINOR = "+"
    MAJOR = "-"

    def to_int(self) -> int:
        """Cast an enum to int."""

        return VersionInt[self.name].value


class VersionInt(IntEnum):
    """An enum used to determine the required version.

    Examples:
        >>> [version.name for version in VersionInt]
        ['NONE', 'PATCH', 'MINOR', 'MAJOR']

        >>> [version.value for version in VersionInt]
        [0, 1, 2, 3]

        >>> [str(version) for version in VersionInt]
        ['', '~', '+', '-']

    .. versionadded:: 1.0.0

    """

    NONE = 0
    PATCH = 1
    MINOR = 2
    MAJOR = 3

    def __str__(self) -> str:
        return VersionStr[self.name].value  # pytype: disable=missing-parameter



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Domain models of :mod:`mantic`.

.. versionadded:: 1.0.0

"""

from ._argument import Argument  # noqa: F401
from ._exit import Exit  # noqa: F401
from ._signature import Signature  # noqa: F401
from ._suffix import Suffix  # noqa: F401
from ._to_value import to_value  # noqa: F401
from ._version import VersionInt, VersionStr  # noqa: F401



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Shared hypothesis strategies.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import Any, Callable, Type

from hypothesis import strategies as st

F = Callable[..., Any]


def layouts(layout: Type[object]) -> st.SearchStrategy[object]:
    """A strategy for layouts."""

    return st.builds(
        layout,
        st.text(),
        name = st.text(),
        size = st.integers(),
        minimum_size = st.integers(),
        ratio = st.integers(),
        visible = st.booleans(),
        )


def panels(panel: Type[object]) -> st.SearchStrategy[object]:
    """A strategy for panels."""

    return st.builds(
        panel,
        st.text(),
        title = st.text(),
        subtitle = st.text(),
        )


def tables(table: Type[object]) -> st.SearchStrategy[object]:
    """A strategy for tables."""

    return st.builds(
        table,
        st.text(),
        )


def signatures(signature: Type[object]) -> st.SearchStrategy[object]:
    """A strategy for signatures."""

    return st.builds(
        signature,
        name = st.text(min_size = 1),
        file = st.text(min_size = 1),
        )


def register(what: Type[object], strategy: F) -> None:
    """Register an hypothesis strategy.

    Args:
        what: What to register a strategy for.
        strategy: The strategy to register.

    Examples:
        >>> from rich.panel import Panel

        >>> register(Panel, panels)
        None

    .. versionadded:: 1.0.0

    """

    st.register_type_strategy(what, strategy)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['mantic',
 'mantic.actions',
 'mantic.domain',
 'mantic.infra',
 'mantic.infra.repo',
 'mantic.utils',
 'mantic_cli',
 'mantic_cli.tasks',
 'mantic_cli.views',
 'mantic_hypothesis']

package_data = \
{'': ['*']}

install_requires = \
['GitPython<4',
 'classes<1',
 'deal<5',
 'invoke<2',
 'numpy>=1.7,<1.22',
 'returns<1',
 'rich<11',
 'termcolor<2',
 'toml<1',
 'typical<3']

extras_require = \
{':python_version >= "3.7" and python_version < "3.8"': ['typing_extensions<4']}

entry_points = \
{'console_scripts': ['mantic = mantic_cli:main.run']}

setup_kwargs = {
    'name': 'mantic',
    'version': '1.1.5',
    'description': 'Command-line tools to facilitate se-mantic versioning.',
    'long_description': '[![Test](https://github.com/maukoquiroga/mantic/workflows/test/badge.svg)](https://github.com/maukoquiroga/mantic/actions?workflow=test)\n[![Type](https://github.com/maukoquiroga/mantic/workflows/type/badge.svg)](https://github.com/maukoquiroga/mantic/actions?workflow=type)\n[![Lint](https://github.com/maukoquiroga/mantic/workflows/lint/badge.svg)](https://github.com/maukoquiroga/mantic/actions?workflow=lint)\n[![Docs](https://github.com/maukoquiroga/mantic/workflows/docs/badge.svg)](https://github.com/maukoquiroga/mantic/actions?workflow=docs)\n[![Docs](https://readthedocs.org/projects/mantic/badge/)](https://mantic.readthedocs.io/)\n[![PyPI](https://img.shields.io/pypi/v/mantic.svg)](https://pypi.org/project/mantic/)\n[![Coverage](https://codecov.io/gh/maukoquiroga/mantic/branch/main/graph/badge.svg)](https://codecov.io/gh/maukoquiroga/mantic)\n\n# Mantic\n\nCommand-line tools to facilitate se-mantic versioning!\n\n## Demo\n\n![mantic](https://user-images.githubusercontent.com/329236/137640522-1673fc7e-8d88-4418-b10a-29e1e4a1408a.gif)\n\n## Installation\n\n```\npip install mantic\n```\n\n## Usage\n\n```\nmantic --help\nmantic --help check-version\nmantic check-version\n```\n\n## License\n\nCopyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>\n\nLicensed under the EUPL-1.2-or-later\nFor details: [https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12](https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)\n',
    'author': 'Mauko Quiroga',
    'author_email': 'mauko@pm.me',
    'maintainer': 'Mauko Quiroga',
    'maintainer_email': 'mauko@pm.me',
    'url': 'https://github.com/maukoquiroga/mantic',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Help view.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import Tuple

import deal
from rich.console import Console
from rich.layout import Layout
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

import mantic_hypothesis as st
from mantic import __name__, utils

from .. import __version__
from ._base import columns, rows
from ._theme import ThemeConsole

st.register(Layout, st.layouts)
st.register(Panel, st.panels)
st.register(Table, st.tables)

_headers = "Flags", "Description", "Default values"
#: Help command headers.

_console = Console()
#: Internal function used to render.


@deal.has("stdout")
@deal.safe
def render(
        command: str,
        doc: str,
        options: Tuple[Tuple[str, ...], ...],
        ) -> None:
    """Render the home view.

    Args:
        command:
            The name of the task.
        doc:
            The documentation of the task.
        options:
            A sequence of tuples containing the commands and their
            descriptions, default values, for each command.

    Examples:
        >>> command = "say-hi!"
        >>> doc = "How to say hi."
        >>> options = [("-f --flag", "How?", "Like that!")]
        >>> render(command, doc, options)
        ...say-hi!...How to say hi...-f --flag...How?...Like that!...

    .. versionadded:: 1.0.0

    """

    content = _content(doc, options)
    main = _main(command, content)
    root = _root(main)

    _console.print(root)


@deal.pure
def _root(main: Panel) -> Layout:
    """Global help container.

    Args:
        main: Main container to wrap.

    Returns:
        Layout: To render.

    Examples:
        >>> content = _content("", ())
        >>> main = _main("command", content)
        >>> _root(main)
        Layout()

    .. versionadded:: 1.0.0

    """

    return columns(rows(main))


@deal.pure
def _main(command: str, content: Table) -> Panel:
    """Main help container.

    Args:
        command: The name of the task.
        content: The inner container.

    Returns:
        The outer panel.

    Examples:
        >>> content = _content("", ())
        >>> main = _main("command", content)
        >>> main.title
        ...command...

    .. versionadded:: 1.0.0

    """

    return Panel(
        content,
        border_style = ThemeConsole.BORDER,
        padding = 5,
        title = _usage(command),
        subtitle = __version__,
        )


@deal.pure
def _content(doc: str, options: Tuple[Tuple[str, ...], ...]) -> Table:
    """Inner usage instructions content.

    Args:
        doc:
            The documentation of the task.
        options:
            A sequence of triples containing the flags, descriptions, and
            default values, for each option.

    Returns:
        The main content.

    Examples:
        >>> doc = "How to say hi."
        >>> options = [("-f --flag", "How?", "Like that!")]
        >>> content = _content(doc, options)
        >>> list(map(lambda option: option._cells, content.columns))
        [['-f --flag'], ['How?'], ['Like that!']]

        >>> content = _content("", [("", "", "")])
        >>> list(map(lambda option: option._cells, content.columns))
        [[''], [''], ['']]

    .. versionadded:: 1.0.0

    """

    table = Table(
        box = None,
        padding = (0, 5, 1, 10),
        row_styles = [ThemeConsole.ROW],
        style = ThemeConsole.HEADER,
        title = f"{doc}\n\n\n",
        )

    list(map(table.add_column, _headers))

    list(utils.dmap(table.add_row, options))

    return table


@deal.pure
def _usage(command: str) -> Text:
    """Usage instructions.

    Args:
        command: The name of the task.

    Returns:
        A formatted usage text.

    Examples:
        >>> _usage("command")
        <text 'Usage: mantic command [--options] …' ...>

    .. versionadded:: 1.0.0

    """

    text = Text(f"Usage: {__name__} {command} [--options] …")

    text.stylize(ThemeConsole.TITLE)

    return text



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Parse task input to options.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

import classes
import deal

from mantic import utils

from ..types import (
    OptionsAlias,
    TupleListAlias,
    TupleListType,
    TupleTupleAlias,
    TupleTupleType,
    )


@classes.typeclass
def to_options(instance) -> OptionsAlias:
    """A task's view content.

    Args:
        instance: Contents to parse.

    Examples:
        >>> example1 = ("-f --flag", ("How?", "Like that!")),
        >>> to_options(example1)
        (('-f --flag', 'How?', 'Like that!'),)

        >>> example2 = ["-f --flag", ("How?", "Like that!")],
        >>> to_options(example2)
        (('-f --flag', 'How?', 'Like that!'),)

        >>> example3 = [("-f --flag", ["How?", "Like that!"])]
        >>> to_options(example3)
        (('-f --flag', 'How?', 'Like that!'),)

        >>> example4 = "-f --flag", ("How?", "Like that!")
        >>> to_options(example4)
        Traceback (most recent call last):
        NotImplementedError: Missing matched typeclass instance for type: tuple

        >>> to_options((*example1, *example2, *example3))
        (('-f --flag', 'How?', 'Like that!'), ('-f --flag', 'How?', 'Like ...))

    .. versionadded:: 1.0.0

    """

    ...  # pytype: disable=bad-return-type


@deal.pure
@to_options.instance(delegate = TupleListType)
def _from_tuple_list(instance: TupleListAlias) -> OptionsAlias:
    return tuple(utils.dcons(instance))


@deal.pure
@to_options.instance(delegate = TupleTupleType)
def _from_tuple_tuple(instance: TupleTupleAlias) -> OptionsAlias:
    return tuple(utils.dcons(instance))



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""A collection od tasks.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from invoke import Collection

from ._check_deprecated import check_deprecated
from ._check_version import check_version


class Tasks(Collection):
    """Interfaces with :mod:`.invoke` to register tasks."""

    def __init__(self) -> None:
        super().__init__()
        self.add_task(check_deprecated)
        self.add_task(check_version)



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Configuration options.

For now, configuration options will be read from any valid ``pyproject.toml``
or ``setup.cfg`` file.

For example::

    [tool.mantic]
    exclude = [".editorconfig", ".gitignore", "tests"]

.. versionchanged:: 1.2.0

.. versionadded:: 1.0.0

"""

from typing import Any, MutableMapping, Sequence, Type

from configparser import ConfigParser
from pathlib import Path

import deal
import toml
import typic


@typic.klass(frozen = True, slots = True, strict = True)
class Config:
    """Provides a configuration representation."""

    ignore: Sequence[str]


@deal.has("import")
def build_config(config: Type[Config]) -> Config:
    """Builds the configuration."""

    from_file: MutableMapping[str, Any]

    if Path("pyproject.toml").exists():
        from_file = toml.load("pyproject.toml")
        return config.transmute(from_file["tool"]["mantic"])

    if Path("setup.cfg").exists():
        parser = ConfigParser()
        parser.read("setup.cfg")
        from_file = {"ignore": parser["tool:mantic"]["ignore"].split()}
        return config.transmute(from_file)

    raise NotImplementedError


config = build_config(Config)



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Suffix schema.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from enum import Enum


class Suffix(str, Enum):
    """An enum to find unique signature names.

    Examples:
        >>> Suffix.BIS
        <Suffix.BIS: '(bis)'>

    .. versionadded:: 1.0.0

    """

    SEMEL = ""
    BIS = "(bis)"
    TER = "(ter)"
    QUATER = "(quater)"
    QUINQUIES = "(quinquies)"
    SEXIES = "(sexies)"
    SEPTIES = "(septies)"
    OCTIES = "(octies)"
    NONIES = "(nonies)"
    DECIES = "(decies)"



# Copyleft (ɔ) 2021 Mauko Quiroga <mauko@pm.me>
#
# Licensed under the EUPL-1.2-or-later
# For details: https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

"""Retrieves file/change information.

.. versionadded:: 1.0.0

"""

from __future__ import annotations

from typing import Tuple

import deal
from git import Repo
from git.exc import GitCommandError


@deal.pre(lambda _: len(_.revision) > 0 and len(_.file) > 0)
@deal.raises(TypeError, ValueError)
@deal.has("io")
def show(revision: str, file: str, repo: str = "") -> str:
    """Retrives the content of a file in a revision.

    Args:
        revision: A commit, a tag, and so on…
        file: The relative file path.
        repo: The git repository path.

    Returns:
        str: The contents of the file.

    Raises:
        TypeError: When arguments are invalid.

    Examples:
        >>> import os
        >>> from pathlib import Path

        >>> repo = Path("./tests/fixtures").resolve()
        >>> source = show("1.0.0", "func.py", str(repo))
        >>> source
        'def function(a, *, b, c, d):\\n    ...'

    .. versionadded:: 1.0.0

    """

    try:
        return (
            Repo(repo)
            .git
            .show(f"{revision}:{file}")
            )
    except GitCommandError as error:
        raise TypeError(error) from error


@deal.pre(lambda _: len(_.revision) > 0)
@deal.raises(TypeError, ValueError)
@deal.has("io")
def tree(revision: str, repo: str = "") -> Tuple[str, ...]:
    """Retrives the list of tracked files in a revision.

    Args:
        revision: A commit, a tag, and so on…
        repo: The git repository path.

    Returns:
        A sequence with the files' names.

    Raises:
        TypeError: When arguments are invalid.

    Examples:
        >>> import os
        >>> from pathlib import Path

        >>> repo = Path("./tests/fixtures").resolve()
        >>> tree("1.0.0", str(repo))
        ('.gitignore', '__init__.py', 'bar.py', 'func.py', 'func_with_chang...)

    .. versionadded:: 1.0.0

    """
    try:
        return tuple(
            Repo(repo)
            .git
            .ls_tree("-r", "--name-only", revision)
            .split()
            )
    except GitCommandError as error:
        raise TypeError(error) from error


@deal.pre(lambda _: len(_.this) > 0 and len(_.that) > 0)
@deal.raises(TypeError, ValueError)
@deal.has("io")
def diff(this: str, that: str, repo: str = "") -> Tuple[str, ...]:
    """Retrives the list of changed files between two revisions.

    Args:
        this: A commit, a tag, and so on…
        that: The same as ``that``, but in the past…
        repo: The git repository path.

    Returns:
        A sequence with the files' names.

    Raises:
        TypeError: When arguments are invalid.

    Examples:
        >>> from pathlib import Path

        >>> repo = Path("./tests/fixtures").resolve()
        >>> diff("2.0.0", "1.0.0", str(repo))
        ('func.py',)

    .. versionadded:: 1.0.0

    """

    try:
        return tuple(
            Repo(repo)
            .git
            .diff("--name-only", f"{that}..{this}")
            .split()
            )
    except GitCommandError as error:
        raise TypeError(error) from error
